"""Tests for getters."""

from napalm.base.test.getters import BaseTestGetters

import pytest


@pytest.mark.usefixtures("set_device_parameters")
class TestGetter(BaseTestGetters):
    """Test get_* methods."""

    def test_method_signatures(self):
        return True



"""Test utils."""

import pytest
from napalm_servertech_pro2 import utils


def test_convert_uptime():
    uptime = "11 days 4 hours 8 minutes 6 seconds"
    assert utils.convert_uptime(uptime) == 965286.0

    uptime = "133 days 1 hour 12 minutes 15 seconds"
    assert isinstance(utils.convert_uptime(uptime), float)

    uptime = "0 days 0 hours 0 minutes 1 second"
    assert utils.convert_uptime(uptime) == 1.0

    with pytest.raises(ValueError):
        utils.convert_uptime("hello")


def test_parse_hardware():
    hardware = "AAAAA (000), 123 MHz, 2048MB RAM, 2048MB FLASH"
    assert utils.parse_hardware(hardware) == {
        "cpu_type": "AAAAA (000)",
        "cpu_freq": 123,
        "ram": 2048,
        "flash": 2048,
    }


def test_validate_actions():
    supported_actions = ["foo", "bar"]
    assert utils.validate_actions("foo", supported_actions) is True

    with pytest.raises(ValueError):
        utils.validate_actions("oof", supported_actions)


def test_server_version():
    headers = {"Server": "ServerTech-AWS/v8.0k"}
    assert utils.server_version(headers) == "8.0k"

    headers = {"Server": "lol"}
    assert utils.server_version(headers) is None

    assert "8.0v" > "8.0m"

    assert "8.0k" < "8.0m"



"""Test fixtures."""

import json
from builtins import super

import pytest
from napalm.base.test import conftest as parent_conftest
from napalm.base.test.double import BaseTestDouble
from requests.models import HTTPError
from napalm_servertech_pro2 import pro2


@pytest.fixture(scope="class")
def set_device_parameters(request):
    """Set up the class."""

    def fin():
        request.cls.device.close()

    request.addfinalizer(fin)

    request.cls.driver = pro2.PRO2Driver
    request.cls.patched_driver = PatchedPRO2Driver
    request.cls.vendor = "servertech"
    parent_conftest.set_device_parameters(request)


def pytest_generate_tests(metafunc):
    """Generate test cases dynamically."""
    parent_conftest.pytest_generate_tests(metafunc, __file__)


class PatchedPRO2Driver(pro2.PRO2Driver):
    """Patched ServerTech PRO2 Driver."""

    def __init__(self, hostname, username, password, timeout=60, optional_args={}):
        """Patched PRO2Driver constructor."""
        super().__init__(hostname, username, password, timeout, optional_args)
        self.patched_attrs = ["api"]

    def close(self):
        pass

    def is_alive(self):
        return {"is_alive": True}

    def open(self):
        self.api = FakePRO2Api()


class FakePRO2Api(BaseTestDouble):
    """ServerTech fake API."""

    REQUESTS = {
        "control/outlets/XX99": {
            "headers": {"Content-Type": "text/html"},
            "status_code": 404,
        }
    }

    def request(self, method, **kwargs):
        address = kwargs["url"].split("/jaws/")[1]
        if self.REQUESTS.get(address):
            return FakeRequest(
                method,
                address,
                self.REQUESTS[address]["headers"],
                self.REQUESTS[address]["status_code"],
            )
        else:
            filename = f"{self.sanitize_text(address)}.json"
            path = self.find_file(filename)
            headers = {
                "Content-Type": "application/json" if method == "GET" else "text/html"
            }
            status_code = 200 if method == "GET" else 204
            return FakeRequest(method, path, headers, status_code)


class FakeRequest:
    """A fake API request."""

    def __init__(self, method, path, headers, status_code):
        self.method = method
        self.path = path
        self.headers = headers
        self.status_code = status_code
        self.text = ""

    def raise_for_status(self):
        if self.status_code >= 400:
            raise HTTPError
        else:
            return True

    def json(self):
        with open(self.path, "r") as file:
            return json.load(file)



"""Tests for getters."""

import pytest
from requests import HTTPError


@pytest.mark.usefixtures("set_device_parameters")
class TestAdd(object):
    """Test additional methods."""

    def test_set_outlet(self):
        res = self.device.set_outlet("AA01", "on")
        assert res["status"] == "success"

        with pytest.raises(ValueError):
            self.device.set_outlet("AA01", "no")

        with pytest.raises(HTTPError):
            self.device.set_outlet("XX99", "on")

    def test_reboot(self):
        res = self.device.restart("normal")
        assert res["status"] == "success"

        with pytest.raises(ValueError):
            self.device.restart("reboot")



"""NAPALM driver for ServerTech PRO2 PDUs"""

import json

import requests
from napalm.base import NetworkDriver
from napalm.base.exceptions import ConnectionException
from netaddr import IPNetwork
from requests.auth import HTTPBasicAuth

from napalm_servertech_pro2.constants import (
    CONFIG_ITEMS,
    LOCAL_USER_LEVELS,
    SUPPORTED_OUTLET_ACTIONS,
    SUPPORTED_RESTART_ACTIONS,
)
from napalm_servertech_pro2.utils import (
    convert_uptime,
    parse_hardware,
    validate_actions,
    server_version,
)


class PRO2Driver(NetworkDriver):
    def __init__(self, hostname, username, password, timeout=60, optional_args=None):
        self.hostname = hostname
        self.username = username
        self.password = password
        self.timeout = timeout
        self.optional_args = optional_args if optional_args else dict()
        self.verify = bool(self.optional_args.get("verify", True))
        self.api = None

        self.platform = "pro2"

    def _req(self, path, method="GET", json=None, raise_err=True):
        url = f"https://{self.hostname}/jaws{path}"
        try:
            req = self.api.request(method, url=url, json=json, verify=self.verify)
            req.raise_for_status()
        except requests.exceptions.HTTPError as err:
            if raise_err:
                raise err
            else:
                return {"err": str(err)}
        if "application/json" in req.headers.get("Content-Type"):
            return req.json()
        else:
            return {
                "status": "success",
                "status_code": req.status_code,
                "content": req.text,
            }

    def open(self):
        """Open a connection to the device."""
        self.api = requests.Session()
        self.api.auth = HTTPBasicAuth(self.username, self.password)
        self.api.verify = self.verify

        try:
            req = self.api.request(
                "GET",
                url=f"https://{self.hostname}/jaws/monitor/system",
                verify=self.verify,
            )
            req.raise_for_status()
        except requests.exceptions.ConnectionError:
            self.api = None
            raise ConnectionException

        version = server_version(req.headers)
        if version and version < "8.0m":
            raise EnvironmentError(
                f"This device is running {version}, while the API was released in 8.0m"
            )

    def close(self):
        self.api.close()

    def is_alive(self):
        return {"is_alive": True if self.api else False}

    def get_config(self, retrieve="all", full=False, sanitized=False):
        if sanitized:
            # There is no way to ask the API not to reveal secrets
            raise NotImplementedError("driver cannot retrieve sanitized configurations")
        config = {}
        for item in CONFIG_ITEMS:
            res = self._req(f"/config/{item}", raise_err=False)
            if "err" not in res:
                config[item] = res
        config_types = {
            "running": json.dumps(config),
            "candidate": "",
            "startup": json.dumps(config),
        }
        if retrieve != "all" and retrieve in config_types:
            return {k: (config_types[k] if k == retrieve else "") for k in config_types}
        else:
            return config_types

    def get_environment(self):
        ret = {k: {} for k in ("fans", "temperature", "power", "cpu", "memory")}

        fans = self._req("/monitor/sensors/fan", raise_err=False)
        if "err" not in fans:
            for fan in fans:
                ret["fans"][fan["name"]] = {
                    "status": True if fans["status"] == "Normal" else False
                }

        temps = self._req("/monitor/sensors/temp", raise_err=False)
        if "err" not in temps:
            for temp in temps:
                if temp["status"] == "Not Found":
                    continue
                config = self._req(f"/config/sensors/temp/{temp['id']}")
                t = temp["temperature_celsius"]
                # thresholds are gathered from configuration
                alert = config["thresholds_celsius"][2]
                critical = config["thresholds_celsius"][3]
                ret["temperature"][temp["name"]] = {
                    "temperature": t,
                    "is_alert": True if t > alert else False,
                    "is_critical": True if t > critical else False,
                }

        cords = self._req("/monitor/cords", raise_err=False)
        for cord in cords:
            utilized = (cord["power_capacity"] / 100) * cord["power_utilized"]
            ret["power"][cord["name"]] = {
                "status": True if cord["status"] == "Normal" else False,
                "capacity": float(cord["power_capacity"]),
                "output": float(utilized),
            }

        system = self._req("/config/info/system", raise_err=False)
        ram = parse_hardware(system["hardware"])["ram"]

        # used RAM can't be retrieved: rather than putting 0 or -1 (inaccurate values),
        # we just assume that it is used at 100%
        ret["memory"] = {
            "available_ram": ram,
            "used_ram": ram,
        }

        return ret

    def get_facts(self):
        info_system = self._req("/config/info/system")
        system = self._req("/config/system")
        network = self._req("/config/network")
        units = self._req("/config/info/units")[0]
        outlets = self._req("/monitor/outlets")

        uptime = convert_uptime(info_system["uptime"])

        return {
            "uptime": uptime,
            "vendor": "ServerTech",  # TODO: try to find a way to get the actual vendor
            "model": units["model_number"],
            "hostname": system["location"],
            "fqdn": network["dhcp_fqdn_name"],
            "os_version": info_system["firmware"],
            "serial_number": units["product_serial_number"],
            "interface_list": [outlet["id"] for outlet in outlets],
        }

    def get_interfaces(self):
        ports = self._req("/monitor/outlets")
        net = self._req("/config/info/network")

        ports.append(
            {
                "id": "NET",
                "name": "management",
                "speed": float(net["speed"].split(" ")[0]),
                "status": "Normal" if net["link"] == "Up" else False,
                "state": "On",
                "mac_address": net["ethernet_mac_address"].replace("-", ":"),
                "mtu": 1500,
            }
        )

        return {
            port["id"]: {
                "is_up": True if port["status"] == "Normal" else False,
                "is_enabled": True if port["state"] == "On" else False,
                "description": port["name"],
                "last_flapped": -1.0,
                "speed": port.get("speed", 0.0),
                "mtu": port.get("mtu", 0),
                "mac_address": port.get("mac_address", ""),
            }
            for port in ports
        }

    def get_interfaces_ip(self):
        net = self._req("/config/info/network")

        ips = {
            4: IPNetwork(f"{net['ipv4_address']}/{net['ipv4_subnet_mask']}"),
            6: IPNetwork(net["ipv6_auto_config_address"]),
        }

        return {
            "net": {
                f"ipv{v}": {str(addr.ip): {"prefix_length": addr.prefixlen}}
                for v, addr in ips.items()
            }
        }

    def get_users(self):
        users = self._req("/config/users/local")

        return {
            user["username"]: {
                "level": LOCAL_USER_LEVELS.get(user["access_level"], 0),
                "password": user["password_secure"],
                "sshkeys": [],
            }
            for user in users
        }

    def set_outlet(self, outlet_id, action):
        """
        Change the status of an outlet

        :param outlet_id: a string
        :param action: a string (values can be: on, off, reboot)
        :return: a dict
        """
        validate_actions(action, SUPPORTED_OUTLET_ACTIONS)

        outlet = self._req(
            f"/control/outlets/{outlet_id}", "PATCH", json={"control_action": action}
        )

        return outlet

    def restart(self, action):
        """
        Restarts the PDU

        :param action: a string (see SUPPORTED_RESTART_ACTIONS for valid values)
        :return: a dict
        """
        validate_actions(action, SUPPORTED_RESTART_ACTIONS)

        restart = self._req("/restart", "PATCH", json={"action": action})

        return restart



import re


def convert_uptime(uptime):
    """Converts a ServerTech uptime string to a number of seconds."""
    time_regex = (
        r"^(?P<days>\d+) days? (?P<hours>\d+) hours?"
        r" (?P<minutes>\d+) minutes? (?P<seconds>\d+) seconds?$"
    )
    m = re.match(
        time_regex,
        uptime,
    )
    if not m:
        raise ValueError("uptime string was not recognized: regex did not match")
    return float(
        int(m.group("days")) * 86400
        + int(m.group("hours")) * 3600
        + int(m.group("minutes")) * 60
        + int(m.group("seconds"))
    )


def parse_hardware(hardware_string):
    """Extract fields from the hardware string."""
    m = re.match(
        r"(?P<cpu_type>[^,;]+), (?P<cpu_freq>\d+) MHz, (?P<ram>\d+)MB RAM, (?P<flash>\d+)MB FLASH",
        hardware_string,
    )
    if not m:
        raise ValueError("hardware string was not recognized: regex did not match")
    return {
        "cpu_type": m.group("cpu_type"),
        "cpu_freq": int(m.group("cpu_freq")),
        "ram": int(m.group("ram")),
        "flash": int(m.group("flash")),
    }


def validate_actions(action, supported_actions):
    """Ensures the inputed action is supported, raises an exception otherwise."""
    if action not in supported_actions:
        raise ValueError(
            f'Action "{action}" is not supported.'
            " the list of valid actions is: {}".format(", ".join(supported_actions))
        )
    return True


def server_version(headers):
    """Extract the firmware version from HTTP headers."""
    version_re = re.compile(r"ServerTech-AWS/v(?P<version>\d+\.\d+\w+)")
    if headers.get("Server"):
        match = version_re.match(headers["Server"])
        if match:
            return match.group("version")



"""Test driver."""

import unittest

from napalm_servertech_pro2 import pro2
from napalm.base.test.base import TestConfigNetworkDriver


class TestConfigPRO2Driver(unittest.TestCase, TestConfigNetworkDriver):
    @classmethod
    def setUpClass(cls):
        """Executed when the class is instantiated."""
        cls.vendor = "servertech"
        cls.device = pro2.PRO2Driver(
            "127.0.0.1",
            "admin",
            "admin",
            timeout=60,
            optional_args={},
        )
        cls.device.open()



"""setup.py file."""

from setuptools import setup, find_packages

with open("requirements.txt", "r") as fs:
    reqs = [r for r in fs.read().splitlines() if (len(r) > 0 and not r.startswith("#"))]

with open("README.md", "r") as fs:
    long_description = fs.read()


__author__ = "Antoine Meillet <antoine.meillet@gmail.com>"

setup(
    name="napalm-servertech-pro2",
    version="0.1.4",
    packages=find_packages(),
    author="Antoine Meillet",
    author_email="antoine.meillet@gmail.com",
    description="NAPALM driver for ServerTech PRO2 PDUs",
    license="Apache 2.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Topic :: Utilities",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
    ],
    url="https://github.com/napalm-automation-community/napalm-servertech-pro2",
    include_package_data=True,
    install_requires=reqs,
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
)



CONFIG_ITEMS = [
    "access",
    "banner",
    "branches",
    "branches-global",
    "cli",
    "cords",
    "cords-global",
    "data-trending",
    "email",
    "ftp",
    "groups",
    "ldap",
    "lines",
    "lines-global",
    "network",
    "outlets",
    "outlets-global",
    "ocps",
    "phases",
    "phases-global",
    "ports",
    "radius",
    "sensors/adc",
    "sensors/adc-global",
    "sensors/contact",
    "sensors/fan",
    "sensors/fan-global",
    "sensors/humidity",
    "sensors/humidity-global",
    "sensors/temp",
    "sensors/temp-global",
    "sensors/water",
    "snmp",
    "sntp",
    "syslog",
    "system",
    "tacacs",
    "units",
    "users/ldap",
    "users/local",
    "users/snmpv3",
    "users/tacacs",
    "web",
    "wlan",
    "ztp",
]

LOCAL_USER_LEVELS = {
    "view only": 10,
    "on only": 11,
    "reboot only": 12,
    "user": 13,
    "power user": 14,
    "admin": 15,
}

SUPPORTED_OUTLET_ACTIONS = ["off", "on", "reboot"]

SUPPORTED_RESTART_ACTIONS = [
    "factory",
    "factory keep network",
    "new firmware",
    "new ssh keys",
    "new x509 certificate",
    "normal",
]
